/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */


/* Copyright (c) 2002-2007, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#define _XOPEN_SOURCE 600

#if defined(_SMP_) || defined(USE_MPD_BASIC) || defined(USE_MPD_RING)
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#endif

/* Required for gethostbyname, which is needed for
 * multi-port, multi-hca */
#include <netdb.h>

#ifdef _SMP_
#ifdef _AFFINITY_
#include <sched.h>
#endif /*_AFFINITY_*/
#endif

#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include "viadev.h"
#include "viaparam.h"
#include "viapriv.h"
#include "viutil.h"

#include "process/pmgr_client.h"
#include "mpid_smpi.h"

#ifdef USE_MPD_RING
static void ib_vapi_bootstrap_ring(int, int, char *, char *, char *,
                                   struct ibv_mr *);
static void ib_vapi_enable_ring(int, int, char *);
#endif

#ifdef _SMP_
#ifdef _AFFINITY_
extern unsigned int viadev_enable_affinity;
#endif
#endif

static void ib_rank_lid_table(int *alladdrs, int len);

#ifdef USE_MPD_RING
static void *barrier_buff;
#endif

static int get_host_id(char *myhostname, int hostname_len);

viadev_info_t viadev;

int vapi_proc_info(int i, char **host, char **exename)
{
    int pid;
    pid = viadev.pids[i];
    *host = viadev.processes[i];
    *exename = viadev.execname;
    return (pid);
}

int MPID_VIA_Init(int *argc, char ***argv, int *size, int *rank)
{
    int local_addr_len;
    int *alladdrs = NULL;
    int *local_addr = NULL;

    int hostid = 0;
    int *allhostids = NULL;
#if (defined (_SMP_) && !((defined (USE_MPD_RING) || defined (USE_MPD_BASIC))))
#ifdef _AFFINITY_ /*_AFFINITY_ start*/
    long N_CPUs_online;

    /*Get number of CPU on machine*/
    if (( N_CPUs_online = sysconf (_SC_NPROCESSORS_ONLN)) < 1 ) {
        perror("sysconf");
    }
#endif /*_AFFINITY_ end*/

#endif

    int mypid;

    /* init the VIA barrier counter */
    viadev.barrier_id = 0;

    /* Device is not yet initialized */
    viadev.initialized = 0;

    /*
     * Find out from the process manager who we are. 
     */

    if (!pmgr_client_init(argc, argv, &viadev.np, &viadev.me,
                          &viadev.global_id, &viadev.processes)) {
        error_abort_all(GEN_EXIT_ERR, "pmgr_client_init failed");
    }

    *size = viadev.np;
    *rank = viadev.me;
    strcpy(viadev.execname, ((const char **) (*argv))[0]);
    viadev.is_finalized = 0;

    /* Now that we know our rank, we should also know
     * our name */
    viadev.my_name = (char *) malloc(sizeof(char) * HOSTNAME_LEN);

    if (NULL == viadev.my_name) {
        error_abort_all(GEN_EXIT_ERR, "malloc failed");
    }

    if (gethostname(viadev.my_name, sizeof(char) * HOSTNAME_LEN) < 0) {
        error_abort_all(GEN_EXIT_ERR, "gethostname failed");
    }

    D_PRINT("size=%d, myrank=%d\n", *size, *rank);


    /* For both SMP and no-SMP, we have the same address length */
    alladdrs = (int *) malloc(3 * viadev.np * sizeof(int));
    local_addr = (int *) malloc(2 * viadev.np * sizeof(int));

    viadev.lid_table = (uint16_t *) malloc(viadev.np * sizeof(int));
    viadev.qp_table = (uint32_t *) malloc(viadev.np * sizeof(int));
    viadev.pids = (int *) malloc(viadev.np * sizeof(int));
    viadev.pids[viadev.me] = mypid = (int) getpid();

    if (alladdrs == NULL || local_addr == NULL ||
        viadev.lid_table == NULL || viadev.qp_table == NULL) {
        error_abort_all(GEN_EXIT_ERR,
                        "malloc for alladdrs/local_addr/lid_table/qp_table");
    }

    hostid = get_host_id(viadev.my_name, HOSTNAME_LEN);
    allhostids= (int *)malloc(viadev.np * sizeof(hostid));
    if (hostid == 0 || allhostids == NULL) {
        error_abort_all(GEN_EXIT_ERR,
                "malloc: in init");
    }
    pmgr_exchange_hostid(hostid, sizeof(hostid), allhostids);
    
    viadev_init_parameters(viadev.np, viadev.me);

  { 

#if defined(USE_MPD_RING) || defined (USE_MPD_BASIC)
    /* 
     * For the scalable network discovery, we create a ring of queue pairs.
     * Then we exchange other queue_pair #'s over the ring. 
     * In future, consider using MPD as a run_time option later. 
     * For now, we stick with some of the process manager code.
     */

    {

        int temp_hostid;
        char *alladdrs_self;
        char *alladdrs_group;
        char *temp_ptr;

#ifdef USE_MPD_RING
        int lhs, rhs;
        char ring_qp_out[64];
        char ring_qp_in[128];

        temp_ptr = (char *) ring_qp_out;
        lhs = (viadev.me + viadev.np - 1) % viadev.np;
        rhs = (viadev.me + 1) % viadev.np;

        /* Prepare my_hca_lid, lhs qp, rsh qp for exchange */
        sprintf(temp_ptr, "%08d", viadev.my_hca_lid);
        temp_ptr += 8;
        sprintf(temp_ptr, "%08d", viadev.boot_qp_hndl[0]->qp_num);
        temp_ptr += 8;
        sprintf(temp_ptr, "%08d", viadev.boot_qp_hndl[1]->qp_num);
        temp_ptr += 8;
        *temp_ptr = '\0';
        D_PRINT("ring_qp_out has %s\n", ring_qp_out);
#endif

        local_addr_len = (viadev.np + 1) * QPLEN_XDR;
        alladdrs_self = (char *) malloc(local_addr_len + 4);
        alladdrs_group = (char *) malloc(viadev.np * local_addr_len + 4);

        /* Prepare all the queue pairs from myself */
        temp_ptr = (char *) alladdrs_self;
        for (i = 0; i < viadev.np; i++) {

            /* Queue pairs for all processes */
            if (i == viadev.me) {
                sprintf(temp_ptr, "%08d", viadev.my_hca_lid);
            } else {
                sprintf(temp_ptr, "%08d", viadev.qp_hndl[i]->qp_num);
            }
            temp_ptr += 8;
        }
        temp_hostid = get_host_id(viadev.my_name, HOSTNAME_LEN);
        sprintf(temp_ptr, "%08X", temp_hostid);
        temp_ptr += 8;
        *temp_ptr = '\0';


#ifdef USE_MPD_BASIC
        {
            int self_length = strlen(alladdrs_self);
            assert(self_length == local_addr_len);
            pmgr_exchange_addresses(alladdrs_self, self_length,
                                    alladdrs_group, viadev.pids);
        }
#else
        D_PRINT("Before exchange, out=%s\n", ring_qp_out);
        /* Save the static buffer address */
        barrier_buff = (void *) alladdrs_group;
        pmgr_exchange_addresses(ring_qp_out, strlen(ring_qp_out),
                                ring_qp_in, viadev.pids);
        D_PRINT("After exchange, in=%s\n", ring_qp_in);

#endif
        D_PRINT("Done exchanging addresses group=%s\n", alladdrs_group);

        /*The memory allocated for alladdrs is not freed at this time. */
        ib_rank_lid_table((int *) alladdrs_group, local_addr_len);

        free(alladdrs_self);
        free(alladdrs_group);
    }
#else                           /* (defined(USE_MPD_BASIC) || defined (USE_MPD_RING)) */

    local_addr_len = viadev.np * sizeof(int);

#ifdef _SMP_
    hostid = get_host_id(viadev.my_name, HOSTNAME_LEN);
    local_addr[viadev.np] = hostid;
    local_addr_len += sizeof(int);

#else                           /* put a useless integer for no-SMP mode */

    local_addr[viadev.np] = viadev.me;
    local_addr_len += sizeof(int);
#endif

    pmgr_exchange_addresses(local_addr, local_addr_len, alladdrs,
                            viadev.pids);
    ib_rank_lid_table(alladdrs, local_addr_len);

    /* initialize connection parameters */
    viadev.connections = (viadev_connection_t *)
        malloc(sizeof(viadev_connection_t) * viadev.np);

    if (NULL == viadev.connections) {
        error_abort_all(GEN_EXIT_ERR,
                        "Unable to malloc connection structures");
    }

    memset(viadev.connections, 0,
           (sizeof(viadev_connection_t) * viadev.np));

#endif                          /* End of ! (defined(USE_MPD_BASIC) || defined (USE_MPD_RING)) */


    free(local_addr);
    free(alladdrs);
    }

    return 0;
}

/*
 * Set up all communication and put VIs in the VI array in viadev
 * The algorithm is as follows (may change when M-VIA supports peer-to-peer
 * connections):
 *   . use client-server routines to establish connections
 *   . for each pair, the lower-numbered process acts as the server
 *   . we should be careful not to establish hotspots. E.g., if
 *     every process connects with others in order 0, 1, 2. We
 *     should also be careful not to serialize. 
 *     there is a nice fancy way to do this, with a communication pattern
 *     essentially the same as in an FFT. 
 *   . However, it is a bit tricky to get right so for now 
 *     do the stupid serialized version (xxx):
 *        for (i = 0; i < me; i++) connect(i)
 *        for (i = me+1; i <= np; i++) accept(i)
 *   
 *  Note that we do not set up a connection to ourselves 
 *
 */

void MPID_Wtime(double *t)
{
    struct timeval tv;
    static int initialized = 0;
    static int sec_base;
    gettimeofday(&tv, NULL);
    if (!initialized) {
        sec_base = tv.tv_sec;
        initialized = 1;
    }
    *t = (double) (tv.tv_sec - sec_base) + (double) tv.tv_usec * 1.0e-6;
}

void MPID_Wtick(double *t)
{
    *t = VIADEV_WTICK;
}

void MPID_VIA_Finalize(void)
{
    /* free connection array space */
    if (viadev.connections) {
        free(viadev.connections);
    }

    /* let the process manager reclaim data structures 
     * before termination */

    pmgr_finalize();

    if (viadev.lid_table) {
        free(viadev.lid_table);
    }
    if (viadev.qp_table) {
        free(viadev.qp_table);
    }
    if (viadev.pids) {
        free(viadev.pids);
    }
    if (viadev.processes) {
        free(viadev.processes);
    }
#ifdef _SMP_
    if (smpi.local_nodes) {
        free(smpi.local_nodes);
    }
    if (smpi.l2g_rank) {
        free(smpi.l2g_rank);
    }
#endif

    D_PRINT("exiting MPID_VIA_Finalize...\n");

}

/* gethostname does return a NULL-terminated string if
 * the hostname length is less than "HOSTNAME_LEN". Otherwise,
 * it is unspecified.
 */
static int get_host_id(char *myhostname, int hostname_len)
{
    int host_id = 0;
    struct hostent *hostent;

    hostent = gethostbyname(myhostname);
    if (hostent == NULL) {
        error_abort_all(GEN_EXIT_ERR,
                        "Null value returned by gethostbyname");
    }

    host_id = ((struct in_addr *) hostent->h_addr_list[0])->s_addr;

    return host_id;
}

static void ib_rank_lid_table(int *alladdrs, int local_addr_len)
{
    int i;

#ifdef _SMP_
    int *hostnames_j = NULL;
    int *smpi_ptr = NULL;
    int j;

#ifdef _AFFINITY_
    long N_CPUs_online;
    unsigned long affinity_mask = 1;
    unsigned long affinity_mask_len = sizeof(affinity_mask);

    /*Get number of CPU on machine */
    if ((N_CPUs_online = sysconf(_SC_NPROCESSORS_ONLN)) < 1) {
        perror("sysconf");
    }
#endif /*_AFFINITY_*/

    hostnames_j = (int *) malloc(viadev.np * sizeof(int));
    smpi.local_nodes = (unsigned int *) malloc(viadev.np * sizeof(int));
    smpi_ptr = (unsigned int *) malloc(viadev.np * sizeof(int));

    if (hostnames_j == NULL || smpi.local_nodes == NULL
        || smpi_ptr == NULL) {
        error_abort_all(GEN_EXIT_ERR,
                        "malloc: in ib_rank_lid_table for SMP");
    }

#endif

#if defined(USE_MPD_BASIC) || defined (USE_MPD_RING)
    for (i = 0; i < viadev.np; i++) {

        char *alladdr_inv;
        char *temp_ptr;
        char temp_lid[12];
        char temp_qp[12];
        char temp_hostid[12];

        alladdr_inv =
            (char *) alladdrs + i * ((viadev.np + 1) * QPLEN_XDR);

        {
            /* Get the qp # */
            temp_ptr = alladdr_inv + viadev.me * QPLEN_XDR;
            strncpy(temp_qp, temp_ptr, QPLEN_XDR);
            temp_qp[QPLEN_XDR] = '\0';

            /* Get hca lid # */
            temp_ptr = alladdr_inv + i * QPLEN_XDR;
            strncpy(temp_lid, temp_ptr, QPLEN_XDR);
            temp_lid[QPLEN_XDR] = '\0';

            /* Get host id # */
            temp_ptr = alladdr_inv + viadev.np * QPLEN_XDR;
            strncpy(temp_hostid, temp_ptr, QPLEN_XDR);
            temp_hostid[QPLEN_XDR] = '\0';
        }

        if (i == viadev.me) {
            /* No QP is allocated for a process itself 
             * If you change this, please change mpirun_rsh.c:684 too */
            viadev.qp_table[i] = -1;
        } else {
            viadev.qp_table[i] = atoi(temp_qp);
        }
        viadev.lid_table[i] = atoi(temp_lid);
        D_PRINT("rank_lid[%d]: hca_lid=%d, qp_num=%d\n",
                i, viadev.lid_table[i], viadev.qp_table[i]);
#ifdef _SMP_
        sscanf(temp_hostid, "%08X", &(hostnames_j[i]));
#endif
    }
#else                           /* (defined(USE_MPD_BASIC) || defined (USE_MPD_RING)) */

    for (i = 0; i < viadev.np; i++) {
        viadev.lid_table[i] = alladdrs[i];
        viadev.qp_table[i] = alladdrs[i + viadev.np];
        D_PRINT("rank_lid[%d]: hca_lid=%d, qp_num=%d\n",
                i, viadev.lid_table[i], viadev.qp_table[i]);

#ifdef _SMP_
        hostnames_j[i] = alladdrs[2 * viadev.np + i];
#endif
    }
#endif                          /* End of ! (defined(USE_MPD_BASIC) || defined (USE_MPD_RING)) */

#ifdef _SMP_

    smpi.only_one_device = 1;
    smpi.num_local_nodes = 0;
    for (j = 0; j < viadev.np; j++) {
        if (hostnames_j[viadev.me] == hostnames_j[j]) {
            if (j == viadev.me) {
                smpi.my_local_id = smpi.num_local_nodes;
#ifdef _AFFINITY_
                if (viadev_enable_affinity > 0) {
                    affinity_mask =
                        1 << (smpi.my_local_id % N_CPUs_online);
                    if (sched_setaffinity
                        (0, affinity_mask_len, &affinity_mask) < 0) {
                        perror("sched_setaffinity");
                    }
                }
#endif                          /* _AFFINITY_ */
            }
            smpi.local_nodes[j] = smpi.num_local_nodes;
            smpi_ptr[smpi.num_local_nodes] = j;
            smpi.num_local_nodes++;
        } else {
            error_abort_all(GEN_EXIT_ERR,
                "Fatal error: To use the ch_smp device all the processes should be on the same node"); 
            smpi.only_one_device = 0;
            smpi.local_nodes[j] = -1;
        }
    }
    smpi.l2g_rank =
        (unsigned int *) malloc(smpi.num_local_nodes * sizeof(int));
    if (smpi.l2g_rank == NULL) {
        error_abort_all(GEN_EXIT_ERR,
                        "malloc: in ib_rank_lid_table for SMP");
    }
    for (j = 0; j < smpi.num_local_nodes; j++) {
        smpi.l2g_rank[j] = smpi_ptr[j];
    }
    free(smpi_ptr);
    free(hostnames_j);
#endif
}

#ifdef PENTIUM_MHZ
double get_us(void)
{
    unsigned long long now;
    __asm __volatile(".byte 0x0f, 0x31":"=A"(now));
    return (double) ((double) now / (PENTIUM_MHZ));
}
#else
double get_us(void)
{
    struct timeval t;
    gettimeofday(&t, NULL);
    return (double) t.tv_sec * (double) 1e6 + (double) t.tv_usec;
}
#endif
